from pathlib import Path

import pandas as pd
import pytest

from vivarium_cluster_tools.psimulate.paths import OutputPaths
from vivarium_cluster_tools.psimulate.results.processing import (
    _concat_metadata,
    _concat_preserve_types,
    write_results_batch,
)


@pytest.mark.parametrize(
    "data_types",
    [
        [0, 1, 2],
        [0, 0.0, 1, 2, 3],
        [0.0, 1.0, 1.987987],
        [0, "a", 1, "b", 1.2],
        ["a", "b", "c", "d"],
    ],
    ids=[
        "just ints",
        "ints and floats",
        "just floats",
        "ints, floats, and strings",
        "just strings",
    ],
)
def test_concat_preserve_types(data_types):
    df = pd.DataFrame([data_types])
    df2 = pd.DataFrame([[d * 2 for d in data_types]])

    result = pd.DataFrame(_concat_preserve_types([df, df2]))

    expected_dtypes = df.dtypes

    expected_shape = (df.shape[0] + df2.shape[0], df.shape[1])

    for c in df:
        assert (result[c] == pd.concat([df[c], df2[c]]).reset_index(drop=True)).all()

    assert result.shape == expected_shape
    assert result.dtypes.sort_index().equals(expected_dtypes)


@pytest.mark.parametrize(
    "data_types",
    [
        [0, 1, 2],
        [0, 0.0, 1, 2, 3],
        [0.0, 1.0, 1.987987],
        [0, "a", 1, "b", 1.2],
        ["a", "b", "c", "d"],
    ],
    ids=[
        "just ints",
        "ints and floats",
        "just floats",
        "ints, floats, and strings",
        "just strings",
    ],
)
def test_concat_results(data_types):
    columns = [chr(i) for i in range(ord("a"), ord("a") + len(data_types))]
    old = pd.DataFrame([data_types], columns=columns)
    new = pd.DataFrame([[d * 2 for d in data_types]], columns=columns)

    old["input_draw"] = old["random_seed"] = 0.0
    new["input_draw"] = new["random_seed"] = 1.0

    combined = _concat_metadata(old, [new])

    expected_dtypes = old.dtypes

    expected_shape = (old.shape[0] + new.shape[0], old.shape[1])

    for c in old:
        assert (combined[c] == pd.concat([old[c], new[c]]).reset_index(drop=True)).all()

    assert combined.shape == expected_shape
    assert combined.dtypes.sort_index().equals(expected_dtypes.sort_index())

    # now no existing results
    no_old_combined = _concat_metadata(pd.DataFrame(), [old, new])

    assert no_old_combined.equals(combined)


def test_write_results_batch(tmp_path):
    output_paths = OutputPaths.from_entry_point_args(
        command="foo",
        input_artifact_path=Path("some/artifact/path"),
        result_directory=tmp_path,
        input_model_spec_path=Path("some/model/spec/path"),
    )
    output_paths.results_dir.mkdir()

    # create and save out already-existing metadata file and results
    existing_metadata_orig = pd.DataFrame({"rows": [1, 2], "batch": [1, 1]})
    existing_results_orig = {
        "results": pd.concat(
            [existing_metadata_orig, pd.DataFrame({"value": [10, 20]})], axis=1
        )
    }
    existing_metadata_orig.to_csv(output_paths.finished_sim_metadata, index=False)
    existing_results_orig["results"].to_parquet(output_paths.results_dir / "results.parquet")

    # create new metadata and results to write (larger than batch size of 2)
    unwritten_metadata_orig = [
        pd.DataFrame({"rows": [3], "batch": [2]}),
        pd.DataFrame({"rows": [4], "batch": [2]}),
        pd.DataFrame({"rows": [5], "batch": [3]}),
    ]
    unwritten_results_orig = []
    batch_size = 2
    for i in range(batch_size + 1):
        unwritten_results_orig += [
            {
                "results": pd.concat(
                    [unwritten_metadata_orig[i], pd.DataFrame({"value": [(i + 3) * 10]})],
                    axis=1,
                )
            }
        ]

    (
        existing_metadata,
        unwritten_metadata,
        existing_results,
        unwritten_results,
    ) = write_results_batch(
        output_paths,
        existing_metadata_orig,
        existing_results_orig,
        unwritten_metadata_orig,
        unwritten_results_orig,
        batch_size,
    )

    # check that updated existing metadata and results have been updated w/ a
    # new batch of 2
    assert existing_metadata.equals(
        pd.concat(
            [existing_metadata_orig] + unwritten_metadata_orig[:batch_size], axis=0
        ).reset_index(drop=True)
    )
    assert existing_results["results"].equals(
        pd.concat(
            [existing_results_orig["results"]]
            + [ur["results"] for ur in unwritten_results_orig[:batch_size]],
            axis=0,
        ).reset_index(drop=True)
    )

    # check that the same update metadata and results have been written out
    assert existing_metadata.equals(pd.read_csv(output_paths.finished_sim_metadata))
    assert existing_results["results"].equals(
        pd.read_parquet(output_paths.results_dir / "results.parquet")
    )

    # check that the leftover unwritten metadata and results are returned
    assert unwritten_metadata == unwritten_metadata_orig[batch_size:]
    assert unwritten_results == unwritten_results_orig[batch_size:]



import pytest
from _pytest.config import Config, argparsing
from _pytest.python import Function


def pytest_addoption(parser: argparsing.Parser) -> None:
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_configure(config: Config) -> None:
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config: Config, items: list[Function]) -> None:
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



# mypy: ignore-errors
"""
===================================
Redis Queue and Registry Management
===================================

Unified interface to multiple Redis Databases.

"""

import random
import time
from collections import defaultdict
from collections.abc import Iterator
from itertools import chain
from typing import Any

import redis
import rq
from loguru import logger
from rq.job import Job
from rq.registry import FinishedJobRegistry, StartedJobRegistry


class QueueManager:
    retries_before_fail = 10
    backoff = 30

    def __init__(self, name: str, connection: redis.Redis):
        self._name = name
        self._logger = logger.bind(queue=name)

        self._queue = rq.Queue("vivarium", connection=connection)
        self._wip = StartedJobRegistry(
            "vivarium", connection=connection, job_class=self._queue.job_class
        )
        self._finished = FinishedJobRegistry(
            "vivarium", connection=connection, job_class=self._queue.job_class
        )

        self._status = {
            "total": 0,
            "pending": 0,
            "running": 0,
            "failed": 0,
            "successful": 0,
            "done": 0.0,
            "workers": 0,
        }
        self._failed = False
        self._completed = False

        self._retries = QueueManager.retries_before_fail
        self._backoff = QueueManager.backoff

    @property
    def name(self) -> str:
        return self._name

    @property
    def failed(self) -> bool:
        return self._failed

    @property
    def completed(self) -> bool:
        return self._completed

    @property
    def jobs_to_finish(self) -> bool:
        return not (self.failed or self.completed)

    def enqueue(self, jobs: list[dict], workhorse_import_path: str) -> None:
        self._logger.info(f"Enqueuing jobs in queue {self.name}")
        for job in jobs:
            # TODO: might be nice to have tighter ttls but it's hard to predict
            #    how long our jobs will take from model to model and the entire
            #    system is short lived anyway
            self._queue.enqueue(
                workhorse_import_path,
                job_parameters=job,
                ttl=60 * 60 * 24 * 2,
                result_ttl=60 * 60,
                job_timeout="7d",
            )

    def get_results(self) -> list:
        self._logger.debug(f"Checking queue {self.name}")
        finished_jobs = self._get_finished_jobs()
        start = time.time()
        results = []
        for job_id in finished_jobs:
            result = self._get_result(job_id)
            if result is not None:
                results.append(result)
        self._logger.debug(
            f"Retrieved {len(results)} results from queue {self.name} in {time.time() - start:.2f}s"
        )
        return results

    def update(self) -> dict[str, int | float]:
        self._update_status()
        return self._status

    def _update_status(self) -> None:
        if self.jobs_to_finish and self._retries > 0:
            self._logger.debug(f"Updating status for queue {self.name}")
            # TODO: Sometimes there are duplicate job_ids, why?
            try:
                # Account for timing discrepancies in accounting and the
                # fact that a job might legit be in more than one of these.
                successful_ids = set(self._finished.get_job_ids()) if self._queue else set()

                running_ids = set(self._wip.get_job_ids()) if self._queue else set()
                # "Successful" for our purposes means the job is done
                # and the results have been retrieved, whereas for rq
                # finished means the job is done and the results have
                # not been retrieved. Once retrieved, jobs disappear
                # from the finished job registry and we'd see all our
                # account eventually just go to zero.
                running_ids = running_ids | successful_ids

                pending_ids = set(self._queue.job_ids) if self._queue else set()
                pending_ids = pending_ids - running_ids

                failed_ids = (
                    set(self._queue.failed_job_registry.get_job_ids())
                    if self._queue
                    else set()
                )
                failed_ids = failed_ids - (running_ids | pending_ids)

                q_workers = len(rq.Worker.all(queue=self._queue)) if self._queue else 0

                self._status["pending"] = len(pending_ids)
                self._status["running"] = len(running_ids) + len(successful_ids)
                self._status["failed"] = len(failed_ids)
                self._status["successful"] = self._status["successful"]
                self._status["total"] = (
                    len(pending_ids)
                    + len(running_ids)
                    + len(failed_ids)
                    + self._status["successful"]
                )
                self._status["workers"] = q_workers
                self._status["done"] = (
                    100 * self._status["successful"] / self._status["total"]
                )

                if len(pending_ids) + len(running_ids) == 0:
                    self._mark_complete()
                elif len(pending_ids) + len(successful_ids) + q_workers == 0:
                    self._logger.info(
                        f"Queue {self.name} ran out of workers with running jobs.  Marking finished."
                    )
                    self._mark_complete()
            except redis.connection.ConnectionError:
                self._sleep_on_it()
                self._update_status()
        else:
            self._mark_failed()

    def _get_finished_jobs(self) -> list[str]:
        if self._retries:
            try:
                return self._finished.get_job_ids()
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                return self._get_finished_jobs()
        else:
            self._mark_failed()
            return []

    def _get_result(self, job_id: str) -> Any:
        job = self._get_job(job_id)
        result = None
        if job is not None:
            start = time.time()
            result = job.result
            end = time.time()
            self._logger.debug(
                f"Deserialized {job_id} result from queue {self.name} in {end - start:.2f}s."
            )
            self._status["successful"] += 1
            self._remove_finished_job(job)
        return result

    def _get_job(self, job_id: str) -> Job | None:
        if self._retries:
            try:
                start = time.time()
                job = self._queue.fetch_job(job_id)
                end = time.time()
                self._logger.debug(
                    f"Fetched job {job_id} from queue {self.name} in {end - start:.2f}s"
                )
                return job
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                return self._get_job(job_id)
        else:
            self._mark_failed()
            return None

    def _remove_finished_job(self, job: Job) -> None:
        if self._retries:
            try:
                self._finished.remove(job)
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                self._remove_finished_job(job)
        else:
            self._mark_failed()

    def _sleep_on_it(self) -> None:
        backoff = self._backoff * random.random()
        self._logger.warning(
            f"Failed to connect to redis for queue {self.name}.  Retrying in {backoff}s."
        )
        self._retries -= 1
        time.sleep(backoff)

    def _mark_failed(self) -> None:
        # TODO: Probably should cleanup redis stuff, but not sure how yet.
        if not (self.failed or self.completed):
            self._logger.warning(
                f"Queue {self.name} has run out of retries.  Marking as failed."
            )
            self._status["failed"] += self._status["pending"] + self._status["running"]
            self._status["pending"] = 0
            self._status["running"] = 0
            self._status["workers"] = 0
            self._failed = True

    def _mark_complete(self) -> None:
        if not (self.failed or self.completed):
            self._logger.info(f"All jobs on queue {self.name} complete.")
            self._status["successful"] += self._status["pending"] + self._status["running"]
            self._status["pending"] = 0
            self._status["running"] = 0
            self._status["workers"] = 0
            self._completed = True


class RegistryManager:
    def __init__(
        self,
        redis_processes: tuple[str, int],
        submitted_workers: int,
        num_already_completed: int,
    ):
        self._logger = logger.bind(queue="all")
        self._logger.info("Building registries.")
        self._queues = [
            QueueManager(str(i), connection=redis.Redis(hostname, port))
            for i, (hostname, port) in enumerate(redis_processes)
        ]
        self._previously_completed = num_already_completed
        self.submitted_workers = submitted_workers

    @property
    def jobs_to_finish(self) -> bool:
        return any([q.jobs_to_finish for q in self._queues])

    def enqueue(self, jobs: list[dict], workhorse_import_path: str) -> None:
        for queue, jobs_chunk in zip(self._queues, self.allocate_jobs(jobs)):
            queue.enqueue(jobs_chunk, workhorse_import_path)

    def allocate_jobs(self, jobs: list[dict]) -> Iterator[list[dict]]:
        """Allocate jobs to queues in a round robin fashion."""
        num_queues = len(self._queues)
        for mod in range(num_queues):
            yield [job for i, job in enumerate(jobs) if i % num_queues == mod]

    def get_results(self) -> list:
        to_check = [q for q in self._queues if q.jobs_to_finish]
        results = []
        for queue in to_check:
            results.extend(queue.get_results())
        return results

    def update_and_report(self) -> dict[str, int | float]:
        status: dict[str, int | float] = defaultdict(int)
        for queue in self._queues:
            queue_status = queue.update()
            for k, v in queue_status.items():
                status[k] += v
        status["done"] = status["successful"] / status["total"] * 100
        status["successful"] += self._previously_completed
        status["unscheduled"] = self.submitted_workers - status["workers"]

        template = (
            "Queue all - Total jobs: {total}, % Done: {done:.2f}% "
            "Pending: {pending}, Running: {running}, Failed: {failed}, Successful: {successful} "
            "Inactive Workers: {unscheduled}."
        )

        self._logger.info(template.format(**status))
        return status

    def get_params_by_job(self) -> dict[str, dict]:
        jobs = list(chain.from_iterable([q._queue.jobs for q in self._queues]))
        return {job.id: job.kwargs["job_parameters"] for job in jobs}

    def __len__(self) -> int:
        return len(self._queues)



# mypy: ignore-errors
"""
=====================================
Reproducible Programming Environments
=====================================

Tools to manage and validate reproducible Python environments for simulation runs.

"""
import os
import subprocess
from pathlib import Path

import click
from loguru import logger


def validate(environment_file: Path) -> None:
    pip_list_proc = subprocess.Popen(
        ["pip", "list", "--format=freeze"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = pip_list_proc.communicate()
    current_environment_list = out.decode().strip().split("\n")

    # Update permissions mask (assign to variable to avoid printing previous value)
    _ = os.umask(0o002)

    if not environment_file.exists():  # original run
        logger.info(f"Writing environment file to {str(environment_file)}.")
        environment_file.write_text("\n".join(current_environment_list))
    else:  # compare with original
        with environment_file.open() as f:
            original_environment_list = [line.replace("\n", "") for line in f]
        current_environment = _convert_pip_list_to_dict(current_environment_list)
        original_environment = _convert_pip_list_to_dict(original_environment_list)
        _compare_environments(current_environment, original_environment)
        logger.info(
            "Validation of environment successful. All pip installed packages match "
            "original versions. Run can proceed."
        )


def _parse_package_version(s: str) -> tuple[str, str]:
    if (
        "no version control" in s
    ):  # installed from non-git controlled source code in editable mode
        s = s.split("(")[1].split(")")[0]  # yields <package>==<version>

    packaged = s.split("==")
    if len(packaged) == 2:  # not installed from source code, e.g., <package>==<version>
        package = packaged[0]
        version = packaged[1]
    else:  # installed from source code in editable mode so more parsing required
        sourced = s.split("=")
        package = sourced[-1]
        version = sourced[0].split("@")[-1].replace("#egg", "")  # version here is git hash
    return package, version


def _convert_pip_list_to_dict(pf_list: list[str]) -> dict:
    return {p: v for p, v in [_parse_package_version(s) for s in pf_list]}


def _compare_environments(current: dict, original: dict) -> None:
    differences = []

    current_packages = set(current.keys())
    original_packages = set(original.keys())

    new_packages = current_packages.difference(original_packages)
    if new_packages:
        differences.append(
            f"The current environment contains the following packages not present "
            f"in the original environment: {new_packages}."
        )

    missing_packages = original_packages.difference(current_packages)
    if missing_packages:
        differences.append(
            f"The current environment is missing the following packages present in the original "
            f"environment: {missing_packages}."
        )

    differing_versions = []
    for p in current_packages.intersection(original_packages):
        if current[p] != original[p]:
            differing_versions.append(f"{p}: {original[p]} -> {current[p]}")

    if differing_versions:
        differences.append(
            f"Different versions found in current and original environment for the following "
            f"packages: {differing_versions}."
        )

    if differences:
        differences = "\n".join(differences)
        logger.info(
            "Differences found between current environment and original environment used for "
            f" this run. \n\nDifferences found are as follows: {differences}. Would you like to proceed?"
        )
        click.confirm(
            "Would you like to proceed psimulate restart with the new environment?",
            abort=True,
        )



# mypy: ignore-errors
"""
==============
psimulate Jobs
==============

"""

from pathlib import Path
from typing import NamedTuple

import numpy as np
import pandas as pd
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate import COMMANDS, branches


class JobParameters(NamedTuple):
    """Parameters for a single distributed simulation job."""

    model_specification: str
    branch_configuration: dict
    input_draw: int
    random_seed: int
    results_path: str
    backup_configuration: dict
    extras: dict

    @property
    def shared(self) -> dict:
        """Parameters shared by all jobs in a psimulate run."""
        return {
            "model_specification": self.model_specification,
            "results_path": self.results_path,
            "backup_configuration": self.backup_configuration,
        }

    @property
    def job_specific(self) -> dict:
        """Parameters that vary by job in a psimulate run."""
        return {
            **self.branch_configuration,
            "input_draw": self.input_draw,
            "random_seed": self.random_seed,
        }

    @property
    def sim_config(self) -> dict:
        """Parameters for the simulation configuration."""
        return {
            **self.branch_configuration,
            "randomness": {
                "random_seed": self.random_seed,
            },
            "input_data": {
                "input_draw_number": self.input_draw,
            },
        }

    def to_dict(self) -> dict:
        # I will never understand why this is a private
        # method of named tuples.
        return self._asdict()

    def __repr__(self) -> str:
        return (
            f"({self.input_draw}, {self.random_seed}, "
            f"{self.model_specification}, {self.branch_configuration})"
        )


def build_job_list(
    command: str,
    model_specification_path: Path,
    output_root: Path,
    keyspace: branches.Keyspace,
    finished_sim_metadata: pd.DataFrame,
    backup_freq: int | None,
    backup_dir: Path,
    backup_metadata_path: Path,
    extras: dict,
) -> tuple[list[JobParameters], int]:
    jobs = []
    number_already_completed = 0

    if command in [COMMANDS.run, COMMANDS.restart, COMMANDS.expand]:
        for input_draw, random_seed, branch_config in keyspace:
            parameters = JobParameters(
                model_specification=str(model_specification_path),
                branch_configuration=branch_config,
                input_draw=int(input_draw),
                random_seed=int(random_seed),
                results_path=str(output_root),
                backup_configuration={
                    "backup_dir": backup_dir,
                    "backup_freq": backup_freq,
                    "backup_metadata_path": backup_metadata_path,
                },
                extras=extras,
            )

            if already_complete(parameters, finished_sim_metadata):
                number_already_completed += 1
            else:
                jobs.append(parameters.to_dict())
    else:
        assert command == COMMANDS.load_test
        for i in range(extras["num_workers"]):
            parameters = JobParameters(
                model_specification=str(model_specification_path),
                branch_configuration={},
                input_draw=0,
                random_seed=i,
                results_path=str(output_root),
                extras={"test_type": extras["test_type"]},
            )
            jobs.append(parameters.to_dict())

    np.random.shuffle(jobs)
    return jobs, number_already_completed


def already_complete(
    job_parameters: JobParameters, finished_sim_metadata: pd.DataFrame
) -> bool:
    if finished_sim_metadata.empty:
        return False

    job_parameter_list = collapse_nested_dict(job_parameters.job_specific)
    mask = pd.Series(True, index=finished_sim_metadata.index)
    for k, v in job_parameter_list:
        if isinstance(v, float):
            mask &= np.isclose(finished_sim_metadata[k], v)
        else:
            mask &= finished_sim_metadata[k] == v
    return np.any(mask)



# mypy: ignore-errors
"""
===============
Vivarium Worker
===============

RQ worker executable for running simulation jobs.

"""

import json
import math
import os
from copy import deepcopy
from pathlib import Path
from time import sleep, time
from traceback import format_exc
from typing import Any

import dill
import pandas as pd
from layered_config_tree import LayeredConfigTree
from loguru import logger
from rq import get_current_job
from vivarium.framework.engine import SimulationContext
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate.environment import ENV_VARIABLES
from vivarium_cluster_tools.psimulate.jobs import JobParameters
from vivarium_cluster_tools.vipin.perf_counters import CounterSnapshot

VIVARIUM_WORK_HORSE_IMPORT_PATH = f"{__name__}.work_horse"


class ParallelSimulationContext(SimulationContext):
    """A parallel simulation context.

    This is identical to ``SimulationContext`` except that it does not write out
    the results to disk in order to allow them to be batch-written.
    """

    def _write_results(self, results: dict[str, pd.DataFrame]) -> None:
        pass


def work_horse(job_parameters: dict) -> tuple[pd.DataFrame, dict[str, pd.DataFrame]]:
    node = f"{ENV_VARIABLES.HOSTNAME.value}"
    job = f"{ENV_VARIABLES.JOB_ID.value}:{ENV_VARIABLES.TASK_ID.value}"

    job_parameters = JobParameters(**job_parameters)

    logger.info(f"Launching new job {job} on {node}")
    logger.info(f"Starting job: {job_parameters}")

    try:
        start_snapshot = CounterSnapshot()
        event = {"start": time()}  # timestamps of application events
        logger.info("Beginning simulation setup.")
        backup = get_backup(job_parameters)
        if backup:
            sim, exec_time = get_sim_from_backup(event, backup)
        else:
            sim, exec_time = initialize_new_sim(event, job_parameters)
        backup_path = run_simulation(job_parameters, event, sim, exec_time)
        results = get_sim_results(sim, job_parameters, start_snapshot, event, exec_time)
        finished_results_metadata = format_and_record_details(job_parameters, results)
        remove_backups(backup_path)

        return finished_results_metadata, results

    except Exception:
        logger.exception("Unhandled exception in worker")
        job = get_current_job()
        job.meta["root_exception"] = format_exc()
        job.save_meta()
        raise
    finally:
        logger.info(f"Exiting job: {job_parameters}")


def get_backup(job_parameters: JobParameters) -> ParallelSimulationContext | None:
    backup_dir = job_parameters.backup_configuration["backup_dir"]
    metadata_path = job_parameters.backup_configuration["backup_metadata_path"]
    try:
        pickle_metadata = pd.read_csv(metadata_path)
        format_val = lambda v: v if isinstance(v, (int, float)) else f'"{v}"'
        query_conditions = " & ".join(
            [
                f"`{k}` == {format_val(v)}"
                for k, v in collapse_nested_dict(job_parameters.job_specific)
            ]
        )

        # Use the query method to find rows that match the lookup parameters
        run_ids = pickle_metadata.query(query_conditions)["job_id"].to_list()
        possible_pickles = [
            Path(backup_dir / run_id).with_suffix(".pkl") for run_id in run_ids
        ]
        existing_pickles = [pickle for pickle in possible_pickles if pickle.exists()]
        if existing_pickles:
            last_pickle = max(existing_pickles, key=os.path.getctime)
            if len(existing_pickles) > 1:
                logger.warning(
                    f"Multiple backups found for {job_parameters}. Using the most recent and deleting the rest."
                )
                for stale_file in set(existing_pickles) - {last_pickle}:
                    os.remove(stale_file)
            with open(last_pickle, "rb") as f:
                sim = dill.load(f)
            current_job_id = get_current_job().id
            logger.info(f"Renaming backup file {last_pickle} to {current_job_id}.pkl")
            if job_parameters.backup_configuration["backup_freq"] is not None:
                # Sleep to prevent FS latency when loading the pickle
                sleep(5)
                os.rename(last_pickle, (backup_dir / str(current_job_id)).with_suffix(".pkl"))
            return sim
    except (OSError, FileNotFoundError):
        logger.info(
            "Missing backup or backup directory. Restarting simulation from beginning."
        )
        return None
    except Exception as e:
        logger.warning(f"Load from backup failed with Exception: {e}")
        logger.warning("Restarting simulation from beginning.")
        return None


def get_sim_from_backup(
    event: dict[str, Any], backup: SimulationContext
) -> tuple[ParallelSimulationContext, dict[str, dict[str, Any]]]:
    logger.info(f"Restarting simulation from saved backup")
    sim = backup
    event["simulation_start"] = time()
    exec_time = {
        "setup_minutes": (event["simulation_start"] - event["start"]) / 60
    }  # execution event
    logger.info(f'Simulation setup completed in {exec_time["setup_minutes"]:.3f} minutes.')
    return sim, exec_time


def initialize_new_sim(
    event: dict[str, Any], job_parameters: JobParameters
) -> tuple[ParallelSimulationContext, dict[str, dict[str, Any]]]:
    sim = ParallelSimulationContext(
        job_parameters.model_specification,
        configuration=job_parameters.sim_config,
        logging_verbosity=job_parameters.extras["sim_verbosity"],
    )
    logger.info("Simulation configuration:")
    logger.info(str(sim.configuration))
    sim.setup()
    event["simulant_initialization_start"] = time()
    exec_time = {
        "setup_minutes": (event["simulant_initialization_start"] - event["start"]) / 60
    }  # execution event
    logger.info(f'Simulation setup completed in {exec_time["setup_minutes"]:.3f} minutes.')

    sim.initialize_simulants()
    event["simulation_start"] = time()
    exec_time["simulant_initialization_minutes"] = (
        event["simulation_start"] - event["simulant_initialization_start"]
    ) / 60
    logger.info(
        f'Simulant initialization completed in {exec_time["simulant_initialization_minutes"]:.3f} minutes.'
    )

    return sim, exec_time


def run_simulation(
    job_parameters: JobParameters,
    event: dict[str, Any],
    sim: ParallelSimulationContext,
    exec_time: dict[str, dict[str, Any]],
) -> Path:
    num_steps = sim.get_number_of_steps_remaining()
    logger.info(f"Starting main simulation loop with {num_steps} time steps")
    backup_path = (
        job_parameters.backup_configuration["backup_dir"] / str(get_current_job().id)
    ).with_suffix(".pkl")
    sim.run(
        backup_freq=job_parameters.backup_configuration["backup_freq"],
        backup_path=backup_path,
    )
    event["results_start"] = time()
    exec_time["main_loop_minutes"] = (event["results_start"] - event["simulation_start"]) / 60
    exec_time["step_mean_seconds"] = (
        event["results_start"] - event["simulation_start"]
    ) / num_steps
    logger.info(
        f'Simulation main loop completed in {exec_time["main_loop_minutes"]:.3f} minutes.'
    )
    logger.info(f'Average step length was {exec_time["step_mean_seconds"]:.3f} seconds.')

    sim.finalize()
    sim.report(print_results=False)
    return backup_path


def get_sim_results(
    sim: ParallelSimulationContext,
    job_parameters: JobParameters,
    start_snapshot: CounterSnapshot,
    event: dict[str, Any],
    exec_time: dict[str, dict[str, Any]],
) -> dict[str, pd.DataFrame]:
    event["end"] = time()
    do_sim_epilogue(start_snapshot, CounterSnapshot(), event, exec_time, job_parameters)
    return sim.get_results()  # Dict[measure, results dataframe]


def do_sim_epilogue(
    start: CounterSnapshot,
    end: CounterSnapshot,
    event: dict,
    exec_time: dict,
    parameters: JobParameters,
) -> None:
    exec_time["results_minutes"] = (event["end"] - event["results_start"]) / 60
    logger.info(f'Results reporting completed in {exec_time["results_minutes"]:.3f} minutes.')
    exec_time["total_minutes"] = (event["end"] - event["start"]) / 60
    logger.info(f'Total simulation run time {exec_time["total_minutes"]:.3f} minutes.')

    perf_log = logger.add(
        Path(ENV_VARIABLES.VIVARIUM_LOGGING_DIRECTORY.value)
        / f"perf.{ENV_VARIABLES.JOB_ID.value}.{ENV_VARIABLES.TASK_ID.value}.log",
        level="DEBUG",
        serialize=True,
    )
    logger.debug(
        json.dumps(
            {
                "host": ENV_VARIABLES.HOSTNAME.value,
                "job_number": ENV_VARIABLES.JOB_ID.value,
                "task_number": ENV_VARIABLES.TASK_ID.value,
                "run_id": get_current_job().id,
                "draw": parameters.input_draw,
                "seed": parameters.random_seed,
                "scenario": parameters.branch_configuration,
                "event": event,
                "exec_time": exec_time,
                "counters": (end - start).to_dict(),
            }
        )
    )
    logger.remove(perf_log)


def format_and_record_details(
    job_parameters: JobParameters, results: dict[str, pd.DataFrame]
) -> pd.DataFrame:
    """Add finished simulation details to results and metadata."""
    finished_results_metadata = pd.DataFrame(index=[0])
    for key, val in collapse_nested_dict(job_parameters.job_specific):
        col_name = key.split(".")[-1]
        for df in results.values():
            # insert the new columns second from the right and use the
            # last part of the key as the column name
            df.insert(df.shape[1] - 1, col_name, val)
        finished_results_metadata[key] = val
    return finished_results_metadata


def remove_backups(backup_path: Path) -> None:
    try:
        os.remove(backup_path)
    except FileNotFoundError:
        pass



import os
import shutil
import time
from pathlib import Path
from subprocess import PIPE, Popen

import pytest

from vivarium_cluster_tools.utilities import backoff_and_retry, mkdir


@pytest.fixture(
    params=[
        ({}, None),
        ({"parents": False}, None),
        ({"parents": True}, None),
        ({"exists_ok": False, "parents": True}, None),
        ({"umask": 0o013, "parents": True}, "drwxrw-r--"),
        ({"parents": True}, None),
        ({"exists_ok": True, "parents": True}, None),
    ]
)
def permissions_params(request):
    return request.param


def test_mkdir_set_permissions(permissions_params: list) -> None:
    # Get prior umask value
    prior_umask = os.umask(0)
    os.umask(prior_umask)

    cwd = Path(os.getcwd())
    parent_dir_name = "parent_dir"
    child_dir_name = "child_dir"

    parent_path = cwd / parent_dir_name
    path = parent_path / child_dir_name

    mkdir_params: dict = permissions_params[0]
    permissions: str | None = permissions_params[1] if permissions_params[1] else "drwxrwxr-x"

    def test_mkdir_permissions():
        mkdir(path, **mkdir_params)
        proc = Popen(
            f"ls -l | grep '{parent_dir_name}' | grep '{permissions}'",
            shell=True,
            stdout=PIPE,
        )
        assert proc.communicate()[0], "Parent directory has incorrect permissions"
        proc = Popen(
            f"ls -l '{parent_dir_name}' | grep '{child_dir_name}' | grep '{permissions}'",
            shell=True,
            stdout=PIPE,
        )
        assert proc.communicate()[0], "Child directory has incorrect permissions"

    try:
        if "parents" not in mkdir_params or not mkdir_params["parents"]:
            with pytest.raises(FileNotFoundError):
                mkdir(path, **mkdir_params)
        else:
            test_mkdir_permissions()

            # Setting new umask doesn't change permissions of existing directories if they exist
            mkdir_params["umask"] = 0o003
            if "exists_ok" not in mkdir_params or not mkdir_params["exists_ok"]:
                with pytest.raises(FileExistsError):
                    mkdir(path, **mkdir_params)
            else:
                test_mkdir_permissions()

    finally:
        if parent_path.exists():
            shutil.rmtree(parent_path)

        assert prior_umask == os.umask(prior_umask), "umask was changed and not reset"


def test_backoff_and_retry():
    class WarningCatcher:
        def __init__(self):
            self.caught_warnings = []

        def warn(self, message, *args, **kwargs):
            self.caught_warnings.append(message)

    wc = WarningCatcher()

    @backoff_and_retry(log_function=wc.warn)
    def successful_function():
        return True

    assert successful_function()
    assert len(wc.caught_warnings) == 0

    wc = WarningCatcher()

    @backoff_and_retry(backoff_seconds=0.1, num_retries=5, log_function=wc.warn)
    def failing_function():
        raise ValueError

    start = time.time()
    with pytest.raises(ValueError):
        failing_function()
    duration = time.time() - start
    # 5 for the retries, 1 for the final fail
    assert len(wc.caught_warnings) == 5 + 1
    assert abs(duration - 0.1 * 5) < 0.01



"""
================
Shared CLI tools
================

"""

from collections.abc import Callable
from pathlib import Path
from typing import TypeVar

import click

# NOTE: The argument type hints for the cli wrappers are not precise; they should
# be type-hinted using Protocols. However, the functions being wrapped are never
# expected to be called in a type-hinted context (because they are used via CLI).
CLIFunction = Callable[..., None]
Decorator = Callable[[CLIFunction], CLIFunction]


def with_verbose_and_pdb(func: CLIFunction) -> CLIFunction:
    func = click.option(
        "-v",
        "verbose",
        count=True,
        help="Configure logging verbosity of main runner for a parallel simulation.",
    )(func)
    func = click.option(
        "--pdb",
        "with_debugger",
        is_flag=True,
        help="Drop into python debugger if an error occurs.",
    )(func)
    return func


def with_sim_verbosity(func: CLIFunction) -> CLIFunction:
    func = click.option(
        "--sim-verbosity",
        "-s",
        type=click.Choice(
            [
                "0",
                "1",
                "2",
            ],
        ),
        required=False,
        default="0",
        show_default=True,
        help="Logging verbosity level of each individual simulation.",
    )(func)
    return func


def coerce_to_full_path(
    ctx: click.Context, param: click.Parameter | None, value: str | None
) -> Path | None:
    if value is not None:
        return Path(value).resolve()
    return None


def pass_shared_options(shared_options: list[Decorator]) -> Decorator:
    """Allows the user to supply a list of click options to apply to a command."""

    def _pass_shared_options(func: CLIFunction) -> CLIFunction:
        # add all the shared options to the command
        for option in shared_options:
            func = option(func)
        return func

    return _pass_shared_options


class MinutesOrNone(click.ParamType):
    """Click param type to allow user to set time in minutes or None."""

    name = "minutesornone"

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> float | None:
        """Converts the value to float seconds from minutes.

        If conversion fails, calls the `fail` method from `click.ParamType`.
        """
        try:
            if value.lower() == "none":
                return None
            # Convert minutes to seconds
            return float(value) * 60.0
        except ValueError:
            self.fail(f"{value!r} is not a valid float or 'none'", param, ctx)


MINUTES_OR_NONE = MinutesOrNone()
